/*
 *  Copyright (C) 2001  Concurrent Computer Corporation
 *  This source is covered by the GNU LGPL Version 2.1
 *  Author: Jason Baietto (jason.baietto@ccur.com)
 *
 *  Abstract the representation of a set of system CPUs.
 *  The current implementation only supports up to 32 CPUs.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define CPUSET_PRIVATE
typedef unsigned long cpuset_t;
#include <cpuset.h>


#define MIN_CPU 0
#define MAX_CPU ((sizeof(cpuset_t)*8)-1)
#define EMPTY_SET 0x0UL
#define FULL_SET  ~(EMPTY_SET)

#define CPUSET_BUFSIZ 1024

void cpuset_init(cpuset_t* setp)
{
   *setp = EMPTY_SET;
}


void cpuset_set_cpu(cpuset_t* setp, cpu_t cpu, int val)
{
   if (!cpuset_is_valid_cpu(cpu)) {
      return;
   }

   if (val) {
      *setp |=  (1ul << cpu);
   } else {
      *setp &= ~(1ul << cpu);
   }
}


int cpuset_get_cpu(cpuset_t* setp, cpu_t cpu)
{
   if (!cpuset_is_valid_cpu(cpu)) {
      return 0;
   }

   if (*setp & (1ul << cpu)) {
      return 1;
   } else {
      return 0;
   }
}


int cpuset_is_empty(cpuset_t* setp)
{
   return (*setp == EMPTY_SET);
}


int cpuset_is_full(cpuset_t* setp)
{
   return (*setp == FULL_SET);
}


int cpuset_is_valid_cpu(cpu_t cpu)
{
   if (cpu < cpuset_min_cpu() || cpu > cpuset_max_cpu()) {
      return 0;
   } else {
      return 1;
   }
}


cpu_t cpuset_min_cpu(void)
{
   return MIN_CPU;
}


cpu_t cpuset_max_cpu(void)
{
   return MAX_CPU;
}


cpuset_t* cpuset_alloc(void)
{
   return (cpuset_t*)malloc(sizeof(cpuset_t));
}


void cpuset_free(cpuset_t* setp)
{
   free(setp);
}


char* cpuset_get_string(cpuset_t* setp)
{
   char buf[CPUSET_BUFSIZ];

   snprintf(buf, CPUSET_BUFSIZ, "%lx", *setp);
   return strdup(buf);
}


void cpuset_set_string(cpuset_t* setp, char* string)
{
   *setp = strtoul(string, NULL, 16);
}


void cpuset_copy(cpuset_t* dst, cpuset_t* src)
{
   *dst = *src;
}


void cpuset_negate(cpuset_t* setp)
{
   *setp = ~(*setp);
}


cpuset_t* cpuset_and(cpuset_t* a, cpuset_t* b)
{
   cpuset_t* c = cpuset_alloc();
   *c = *a & *b;
   return c;
}


void cpuset_and_eq(cpuset_t* a, cpuset_t* b)
{
   *a &= *b;
}


cpuset_t* cpuset_or(cpuset_t* a, cpuset_t* b)
{
   cpuset_t* c = cpuset_alloc();
   *c = *a | *b;
   return c;
}


void cpuset_or_eq(cpuset_t* a, cpuset_t* b)
{
   *a |= *b;
}


cpuset_t* cpuset_xor(cpuset_t* a, cpuset_t* b)
{
   cpuset_t* c = cpuset_alloc();
   *c = *a ^ *b;
   return c;
}


void cpuset_xor_eq(cpuset_t* a, cpuset_t* b)
{
   *a ^= *b;
}


int cpuset_count(cpuset_t* setp)
{
   cpu_t min_cpu = cpuset_min_cpu();
   cpu_t max_cpu = cpuset_max_cpu();
   int count = 0;
   cpu_t i;

   for (i = min_cpu; i <= max_cpu; i++) {
      if (cpuset_get_cpu(setp, i)) {
         count++;
      }
   }

   return count;
}


int cpuset_is_eq(cpuset_t* a, cpuset_t* b)
{
   return (*a == *b);
}

/* deprecated */

#ifdef cpuset_eq
#undef cpuset_eq
#endif
int cpuset_eq(cpuset_t*, cpuset_t*);
int cpuset_eq(cpuset_t* a, cpuset_t* b)
{
   return cpuset_is_eq(a, b);
}
