/*
 *  Copyright (C) 2001  Concurrent Computer Corporation
 *  This source is covered by the GNU LGPL Version 2.1.
 *  Author: Jason Baietto (jason.baietto@ccur.com)
 *
 *  Implement the mpadvise(2) "system call" by messing with
 *  /proc files.  In the future mpadvise may become an actual
 *  system call (or a thin layer on top of other system calls).
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAS_CPUSTAT_SYSCALL
#include <sys/cpustat.h>
#endif
#include <fcntl.h>
#include <dirent.h>
#include <mpadvise.h>
#include <proc_walk.h>
#include <proc_stat.h>
#include <sched.h>
#include <malloc.h>
#include <errno.h>
#include <asm/unistd.h>


#define MPA_PATH_BUFSIZ 1024
#define MPA_BIAS_BUFSIZ 1024
#define MPA_AFFINITY_FILE "affinity"


/*
 *  Hard code this until I figure out a way to determine this.
 */
#define BOOT_CPU 0

/*
 *  Provide a local gettid entry point.  On systems that do not support
 *  gettid only process level functionality will be provided.
 */
static pid_t gettid(void) {
   pid_t tid;
#ifdef __NR_gettid
   tid = (pid_t)syscall(__NR_gettid);
   if (errno == ENOSYS)
      /* fall through */
#endif
      tid = getpid();
   return tid;
}

/***********************************************************************
 *                                                                     *
 *  Functions to implement MPA_PRC_GETBIAS AND MPA_PRC_SETBIAS         *
 *                                                                     *
 ***********************************************************************/


#ifdef WITH_PROC_AFFINITY


static FILE* open_proc_bias_set(pid_t id, char* mode)
{
   FILE* bias_file;
   char buf[MPA_PATH_BUFSIZ];
   int ret;

   ret = snprintf(buf, MPA_PATH_BUFSIZ, "/proc/%d/" MPA_AFFINITY_FILE, id);
   if (ret <= 0 || ret >= MPA_PATH_BUFSIZ) {
      return NULL;
   }
   bias_file = fopen(buf, mode);
   return bias_file;
}


static void close_proc_bias_set(FILE* bias_file)
{
   if (bias_file) {
      fclose(bias_file);
   }
}


static int read_proc_bias_set(FILE* bias_file, cpuset_t* setp)
{
   char buf[MPA_BIAS_BUFSIZ];
   int ret;

   ret = fseek(bias_file, 0, SEEK_SET);
   if (ret != 0) {
      return MPA_FAILURE;
   }

   ret = fread(buf, sizeof(char), MPA_BIAS_BUFSIZ, bias_file);
   if (ret == 0 || ret >= MPA_BIAS_BUFSIZ) {
      return MPA_FAILURE;
   }
   buf[ret+1] = '\0';

   cpuset_set_string(setp, buf);
   if (cpuset_is_empty(setp)) {
      errno = EIO;
      return MPA_FAILURE;
   }

   return MPA_SUCCESS;
}


static int write_proc_bias_set(FILE* bias_file, cpuset_t* setp)
{
   char* string;
   int ret;
   int len;

   if (cpuset_is_empty(setp)) {
      errno = EINVAL;
      return MPA_FAILURE;
   }

   ret = fseek(bias_file, 0, SEEK_SET);
   if (ret != 0) {
      return MPA_FAILURE;
   }

   string = cpuset_get_string(setp);
   if (!string || string[0] == 0) {
      errno = EIO;
      return MPA_FAILURE;
   }
   len = strlen(string);

   ret = fwrite(string, sizeof(char), len, bias_file);
   free(string);
   if (ret != len) {
      return MPA_FAILURE;
   }

   return MPA_SUCCESS;
}


static int get_proc_bias_set(pid_t id, cpuset_t* setp)
{
   FILE* bias_file;
   int ret;

   bias_file = open_proc_bias_set(id, "r");
   if (bias_file == NULL) {
      if (errno == ENOENT) {
         errno = ESRCH;
      }
      return MPA_FAILURE;
   }
	
   ret = read_proc_bias_set(bias_file, setp);
   close_proc_bias_set(bias_file);

   return ret;
}


static int set_proc_bias_set(pid_t id, cpuset_t* setp)
{
   FILE* bias_file;
   int ret;

   bias_file = open_proc_bias_set(id, "w");
   if (bias_file == NULL) {
      if (errno == ENOENT) {
         errno = ESRCH;
      }
      return MPA_FAILURE;
   }
	
   ret = write_proc_bias_set(bias_file, setp);
   close_proc_bias_set(bias_file);

   return ret;
}

#elif defined(WITH_SCHED_AFFINITY)

#ifdef CPU_SETSIZE

/*
 *  Assume new-style sched_{get|set}affinity interface with mask
 *  manipulation macros.
 */

static int get_proc_bias_set(pid_t id, cpuset_t* setp)
{
   cpu_set_t set;
   int len = cpuset_max_cpu();

#if (NUM_SCHED_AFFINITY_PARAMS == 2)
   int ret = sched_getaffinity(id, &set);
#elif (NUM_SCHED_AFFINITY_PARAMS == 3)
   int ret = sched_getaffinity(id, sizeof(set), &set);
#else
#error "Unsupported sched_getaffinity API"
#endif
   int cpu;

   cpuset_init(setp);
   if (!ret) {
      for (cpu = cpuset_min_cpu(); cpu < len; cpu++) {
         if (CPU_ISSET(cpu, &set)) {
            cpuset_set_cpu(setp, cpu, 1);
         }
      }
   }
   return ret;
}

static int set_proc_bias_set(pid_t id, cpuset_t* setp)
{
   cpu_set_t set;
   int len = cpuset_max_cpu();
   int ret;
   int cpu;

   CPU_ZERO(&set);
   for (cpu = cpuset_min_cpu(); cpu < len; cpu++) {
      if (cpuset_get_cpu(setp, cpu)) {
         CPU_SET(cpu, &set);
      }
   }
#if (NUM_SCHED_AFFINITY_PARAMS == 2)
   ret = sched_setaffinity(id, &set);
#elif (NUM_SCHED_AFFINITY_PARAMS == 3)
   ret = sched_setaffinity(id, sizeof(set), &set);
#else
#error "Unsupported sched_setaffinity API"
#endif
   return ret;
}

#else /* !CPU_SETSIZE */

/*
 *  Assume old-style sched_{get|set}affinity interface (3 parameters)
 *  without mask manipulation macros (assume "ulong" sized mask).
 */

static int get_proc_bias_set(pid_t id, cpuset_t* setp)
{
   unsigned long mask;
   int ret;
   int cpu;

   ret = sched_getaffinity(id, sizeof(mask), &mask);
   if (ret == -1) {
      return MPA_FAILURE;
   }

   cpuset_init(setp);
   for (cpu = cpuset_min_cpu(); cpu < sizeof(unsigned long)*8; cpu++) {
      if (mask & (1 << cpu)) {
         cpuset_set_cpu(setp, cpu, 1);
      }
   }

   return MPA_SUCCESS;
}


static int set_proc_bias_set(pid_t id, cpuset_t* setp)
{
   unsigned long mask;
   int ret;
   int cpu;

   mask = 0;
   for (cpu = cpuset_min_cpu(); cpu < sizeof(unsigned long)*8; cpu++) {
      if (cpuset_get_cpu(setp, cpu) == 1) {
         mask |= (1 << cpu);
      }
   }

   ret = sched_setaffinity(id, sizeof(mask), &mask);
   if (ret == -1) {
      return MPA_FAILURE;
   }

   return MPA_SUCCESS;
}

#endif /* !CPU_SETSIZE */

#else
#error "Must define either WITH_SCHED_AFFINITY or WITH_PROC_AFFINITY"
#endif


/***********************************************************************
 *                                                                     *
 *  Functions to implement MPA_CPU_ACTIVE                              *
 *                                                                     *
 ***********************************************************************/

#ifdef HAS_CPUSTAT_SYSCALL

static int get_active_cpu_set(cpuset_t* setp)
{
   int ret;
   int cpu;
   void* avail_mask;
   uint mask_size = 0;

   cpuset_init(setp);

   ret = cpustat(CPUSTAT_GET_CPUMASK_SIZE, sizeof(mask_size), &mask_size);
   if (ret != 0 || mask_size == 0) {
      if (errno == ENOSYS) {
         /* try to fall back to sysconf */
         int cpu_count = sysconf(_SC_NPROCESSORS_ONLN);
         if (cpu_count <= 0) {
            return MPA_FAILURE;
         } else {
            for (cpu=0; cpu < cpu_count; cpu++) {
               cpuset_set_cpu(setp, cpu, 1);
            }
            return MPA_SUCCESS;
         }
      } else {
         return MPA_FAILURE;
      }
   }

   avail_mask = (void*)malloc(mask_size);
   if (avail_mask == NULL) {
      return MPA_FAILURE;
   }

   ret = cpustat(CPUSTAT_GET_AVAIL_MASK, mask_size, avail_mask);
   if (ret != 0) {
      ret = MPA_FAILURE;
   } else {
      for (cpu=0; cpu < (int)mask_size * 8; cpu++) {
         if (CPU_ISSET(cpu, (cpu_set_t*)avail_mask)) {
            cpuset_set_cpu(setp, cpu, 1);
         }
      }
      if (cpuset_is_empty(setp)) {
         ret = MPA_FAILURE;
      } else {
         ret = MPA_SUCCESS;
      }
   }

   free(avail_mask);
   return ret;
}


static int get_present_cpu_set(cpuset_t* setp)
{
   int ret;
   int cpu;
   void* online_mask = NULL;
   uint mask_size = 0;

   cpuset_init(setp);

   ret = cpustat(CPUSTAT_GET_CPUMASK_SIZE, sizeof(mask_size), &mask_size);
   if (ret != 0 || mask_size == 0) {
      if (errno == ENOSYS) {
         /* try to fall back to sysconf */
         int cpu_count = sysconf(_SC_NPROCESSORS_CONF);
         if (cpu_count <= 0) {
            return MPA_FAILURE;
         } else {
            for (cpu=0; cpu < cpu_count; cpu++) {
               cpuset_set_cpu(setp, cpu, 1);
            }
            return MPA_SUCCESS;
        }
      } else {
         return MPA_FAILURE;
      }
   }

   online_mask = (void*)malloc(mask_size);
   if (online_mask == NULL) {
      return MPA_FAILURE;
   }

   ret = cpustat(CPUSTAT_GET_ONLINE_MASK, mask_size, online_mask);
   if (ret != 0) {
      ret = MPA_FAILURE;
   } else {
      for (cpu=0; cpu < (int)mask_size * 8; cpu++) {
         if (CPU_ISSET(cpu, (cpu_set_t*)online_mask)) {
            cpuset_set_cpu(setp, cpu, 1);
         }
      }
      if (cpuset_is_empty(setp)) {
         ret = MPA_FAILURE;
      } else {
         ret = MPA_SUCCESS;
      }
   }

   free(online_mask);
   return ret;
}

#else /* !HAS_CPUSTAT_SYSCALL */

static int get_active_cpu_set(cpuset_t* setp)
{
   int cpu;
   int cpu_count;

   cpuset_init(setp);
   cpu_count = sysconf(_SC_NPROCESSORS_ONLN);
   if (cpu_count <= 0) {
      return MPA_FAILURE;
   }
   for (cpu=0; cpu < cpu_count; cpu++) {
      cpuset_set_cpu(setp, cpu, 1);
   }
   return MPA_SUCCESS;
}


static int get_present_cpu_set(cpuset_t* setp)
{
   int cpu;
   int cpu_count;

   cpuset_init(setp);
   cpu_count = sysconf(_SC_NPROCESSORS_CONF);
   if (cpu_count <= 0) {
      return MPA_FAILURE;
   }
   for (cpu=0; cpu < cpu_count; cpu++) {
      cpuset_set_cpu(setp, cpu, 1);
   }
   return MPA_SUCCESS;
}

#endif


static int get_boot_cpu_set(cpuset_t* setp)
{
   cpuset_init(setp);
   cpuset_set_cpu(setp, BOOT_CPU, 1);
   return MPA_SUCCESS;
}


/***********************************************************************
 *                                                                     *
 *  The mpadvise(2) external entry point                               *
 *                                                                     *
 ***********************************************************************/


int mpadvise(int cmd, int which, int who, cpuset_t* setp)
{
   int ret;
   int success = 0; /* successful operation count */

   switch (cmd) {
      case MPA_PRC_GETBIAS: {
         switch (which) {
            case MPA_TID: {
               if (who == 0) {
                  who = gettid();
               }
               ret = get_proc_bias_set(who, setp);
               if (ret == MPA_SUCCESS) {
                  success++;
               }
               break;
            }
            case MPA_PID: {
               pid_t tid;
               twid_t* twid;
               cpuset_t* tmp;

               if (who == 0) {
                  who = getpid();
               }
               tmp = cpuset_alloc();
               if (tmp == NULL) {
                  return MPA_FAILURE;
               }
               twid = thread_walk_init(who);
               if (twid == NULL) {
                  cpuset_free(tmp);
                  return MPA_FAILURE;
               }
               cpuset_init(setp);
               while ((tid = thread_walk_get_next_tid(twid))) {
                  ret = get_proc_bias_set(tid, tmp);
                  if (ret == MPA_SUCCESS) {
                     cpuset_or_eq(setp, tmp);
                     success++;
                  }
               }
               cpuset_free(tmp);
               thread_walk_done(twid);
               break;
            }
            case MPA_PGID: {
               pid_t tid;
               pwid_t* pwid;
               pid_t pgrp;
               cpuset_t* tmp;

               if (who == 0) {
                  who = getpgrp();
               }
               tmp = cpuset_alloc();
               if (tmp == NULL) {
                  return MPA_FAILURE;
               }
               pwid = proc_walk_init();
               if (pwid == NULL) {
                  cpuset_free(tmp);
                  return MPA_FAILURE;
               }
               cpuset_init(setp);
               while ((tid = proc_walk_get_next_tid(pwid))) {
                  ret = proc_stat(tid, PROC_STAT_PGRP, &pgrp);
                  if (ret == PROC_STAT_FAILURE) {
                     /* ignore as /proc is not atomic */
                     continue;
                  }
                  if (pgrp == who) {
                     ret = get_proc_bias_set(tid, tmp);
                     if (ret == MPA_SUCCESS) {
                        cpuset_or_eq(setp, tmp);
                        success++;
                     }
                  }
               }
               cpuset_free(tmp);
               proc_walk_done(pwid);
               break;
            }
            case MPA_UID: {
               pid_t tid;
               pwid_t* pwid;
               uid_t uid;
               cpuset_t* tmp;

               if (who == 0) {
                  who = geteuid();
               }
               tmp = cpuset_alloc();
               if (tmp == NULL) {
                  return MPA_FAILURE;
               }
               pwid = proc_walk_init();
               if (pwid == NULL) {
                  cpuset_free(tmp);
                  return MPA_FAILURE;
               }
               cpuset_init(setp);
               while ((tid = proc_walk_get_next_tid(pwid))) {
                  ret = proc_stat(tid, PROC_STAT_UID, &uid);
                  if (ret == PROC_STAT_FAILURE) {
                     /* ignore as /proc is not atomic */
                     continue;
                  }
                  if (uid == (uid_t)who) {
                     ret = get_proc_bias_set(tid, tmp);
                     if (ret == MPA_SUCCESS) {
                        cpuset_or_eq(setp, tmp);
                        success++;
                     }
                  }
               }
               cpuset_free(tmp);
               proc_walk_done(pwid);
               break;
            }
            default: {
               errno = EINVAL;
               return MPA_FAILURE;
               break;
            }
         }
         break;
      }

      case MPA_PRC_SETBIAS: {
         switch (which) {
            case MPA_TID: {
               if (who == 0) {
                  who = gettid();
               }
               ret = set_proc_bias_set(who, setp);
               if (ret == MPA_SUCCESS) {
                  success++;
               }
               break;
            }
            case MPA_PID: {
               pid_t tid;
               twid_t* twid;

               if (who == 0) {
                  who = getpid();
               }
               twid = thread_walk_init(who);
               if (twid == NULL) {
                  return MPA_FAILURE;
               }
               while ((tid = thread_walk_get_next_tid(twid))) {
                  ret = set_proc_bias_set(tid, setp);
                  if (ret == MPA_SUCCESS) {
                     success++;
                  }
               }
               thread_walk_done(twid);
               break;
            }
            case MPA_PGID: {
               pid_t tid;
               pwid_t* pwid;
               pid_t pgrp;

               if (who == 0) {
                  who = getpgrp();
               }
               pwid = proc_walk_init();
               if (pwid == NULL) {
                  return MPA_FAILURE;
               }
               while ((tid = proc_walk_get_next_tid(pwid))) {
                  ret = proc_stat(tid, PROC_STAT_PGRP, &pgrp);
                  if (ret == PROC_STAT_FAILURE) {
                     /* ignore as /proc is not atomic */
                     continue;
                  }
                  if (pgrp == who) {
                     ret = set_proc_bias_set(tid, setp);
                     if (ret == MPA_SUCCESS) {
                        success++;
                     }
                  }
               }
               proc_walk_done(pwid);
               break;
            }
            case MPA_UID: {
               pid_t tid;
               pwid_t* pwid;
               uid_t uid;

               if (who == 0) {
                  who = geteuid();
               }
               pwid = proc_walk_init();
               if (pwid == NULL) {
                  break;
               }
               while ((tid = proc_walk_get_next_tid(pwid))) {
                  ret = proc_stat(tid, PROC_STAT_UID, &uid);
                  if (ret == PROC_STAT_FAILURE) {
                     /* ignore as /proc is not atomic */
                     continue;
                  }
                  if (uid == (uid_t)who) {
                     ret = set_proc_bias_set(tid, setp);
                     if (ret == MPA_SUCCESS) {
                        success++;
                     }
                  }
               }
               proc_walk_done(pwid);
               break;
            }
            default: {
               errno = EINVAL;
               return MPA_FAILURE;
               break;
            }
         }
         break;
      }

      case MPA_CPU_ACTIVE: {
         ret = get_active_cpu_set(setp);
         if (ret == MPA_SUCCESS) {
            success++;
         }
         break;
      }

      case MPA_CPU_PRESENT: {
         ret = get_present_cpu_set(setp);
         if (ret == MPA_SUCCESS) {
            success++;
         }
         break;
      }

      case MPA_PRC_GETRUN: {
         switch (which) {
            case MPA_TID: {
               cpu_t cpu;

               if (who == 0) {
                  who = gettid();
               }
               cpuset_init(setp);
               ret = proc_stat(who, PROC_STAT_PROCESSOR, &cpu);
               if (ret == PROC_STAT_SUCCESS) {
                  cpuset_set_cpu(setp, cpu, 1);
                  success++;
               }
               break;
            }
            case MPA_PID: {
               pid_t tid;
               cpu_t cpu;
               twid_t* twid;

               if (who == 0) {
                  who = getpid();
               }
               twid = thread_walk_init(who);
               if (twid == NULL) {
                  return MPA_FAILURE;
               }
               cpuset_init(setp);
               while ((tid = thread_walk_get_next_tid(twid))) {
                  ret = proc_stat(tid, PROC_STAT_PROCESSOR, &cpu);
                  if (ret == PROC_STAT_SUCCESS) {
                     cpuset_set_cpu(setp, cpu, 1);
                     success++;
                  }
               }
               thread_walk_done(twid);
               break;
            }
            case MPA_PGID: {
               pid_t tid;
               pwid_t* pwid;
               pid_t pgrp;
               cpu_t cpu;

               if (who == 0) {
                  who = geteuid();
               }
               pwid = proc_walk_init();
               if (pwid == NULL) {
                  return MPA_FAILURE;
               }
               cpuset_init(setp);
               while ((tid = proc_walk_get_next_tid(pwid))) {
                  ret = proc_stat(tid, PROC_STAT_PGRP, &pgrp);
                  if (ret == PROC_STAT_FAILURE) {
                     /* ignore as /proc is not atomic */
                     continue;
                  }
                  if (pgrp == who) {
                     ret = proc_stat(tid, PROC_STAT_PROCESSOR, &cpu);
                     if (ret == PROC_STAT_SUCCESS) {
                        cpuset_set_cpu(setp, cpu, 1);
                        success++;
                     }
                  }
               }
               proc_walk_done(pwid);
               break;
            }
            case MPA_UID: {
               pid_t tid;
               pwid_t* pwid;
               uid_t uid;
               cpu_t cpu;

               if (who == 0) {
                  who = getpgrp();
               }
               pwid = proc_walk_init();
               if (pwid == NULL) {
                  return MPA_FAILURE;
               }
               cpuset_init(setp);
               while ((tid = proc_walk_get_next_tid(pwid))) {
                  ret = proc_stat(tid, PROC_STAT_UID, &uid);
                  if (ret == PROC_STAT_FAILURE) {
                     /* ignore as /proc is not atomic */
                     continue;
                  }
                  if (uid == (uid_t)who) {
                     ret = proc_stat(tid, PROC_STAT_PROCESSOR, &cpu);
                     if (ret == PROC_STAT_SUCCESS) {
                        cpuset_set_cpu(setp, cpu, 1);
                        success++;
                     }
                  }
               }
               proc_walk_done(pwid);
               break;
            }
            default: {
               errno = EINVAL;
               return MPA_FAILURE;
               break;
            }
         }
         break;
      }

      case MPA_CPU_BOOT: {
         ret = get_boot_cpu_set(setp);
         if (ret == MPA_SUCCESS) {
            success++;
         }
         break;
      }

      default: {
         errno = EINVAL;
         return MPA_FAILURE;
         break;
      }
   }


   if (success == 0) {
      return MPA_FAILURE;
   } else {
      return cpuset_count(setp);
   }
}
