/*
 *  Copyright (C) 2001  Concurrent Computer Corporation
 *  This source is covered by the GNU LGPL Version 2.1
 *  Author: Jason Baietto (jason.baietto@ccur.com)
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <dirent.h>
#include <string.h>
#include <proc_stat.h>
#include <libgen.h>

#define PROC_PATH_BUFSIZ 1024
#define PROC_STAT_BUFSIZ 1024
#define PROC_STATUS_BUFSIZ 4096
#define PROC_CMDLINE_BUFSIZ 4096

static int read_line(FILE*, char*, int);


static FILE* open_proc_stat(pid_t id)
{
   FILE* stat_file;
   char buf[PROC_PATH_BUFSIZ];
   int ret;

   ret = snprintf(buf, PROC_PATH_BUFSIZ, "/proc/%d/stat", id);
   if (ret <= 0 || ret >= PROC_PATH_BUFSIZ) {
      return NULL;
   }
   stat_file = fopen(buf, "r");
   return stat_file;
}


static FILE* open_proc_status(pid_t id)
{
   FILE* status_file;
   char buf[PROC_PATH_BUFSIZ];
   int ret;

   ret = snprintf(buf, PROC_PATH_BUFSIZ, "/proc/%d/status", id);
   if (ret <= 0 || ret >= PROC_PATH_BUFSIZ) {
      return NULL;
   }
   status_file = fopen(buf, "r");
   return status_file;
}


static FILE* open_proc_cmdline(pid_t id)
{
   FILE* cmdline_file;
   char buf[PROC_PATH_BUFSIZ];
   int ret;

   ret = snprintf(buf, PROC_PATH_BUFSIZ, "/proc/%d/cmdline", id);
   if (ret <= 0 || ret >= PROC_PATH_BUFSIZ) {
      return NULL;
   }
   cmdline_file = fopen(buf, "r");
   return cmdline_file;
}


static void close_proc_stat(FILE* stat_file)
{
   if (stat_file != NULL) {
      fclose(stat_file);
   }
}


static void close_proc_status(FILE* status_file)
{
   if (status_file != NULL) {
      fclose(status_file);
   }
}


static void close_proc_cmdline(FILE* cmdline_file)
{
   if (cmdline_file != NULL) {
      fclose(cmdline_file);
   }
}



/*
 *  Note that PROC_STAT_PID and PROC_STAT_UID are not handled here.
 */
static int extract_stat(char* line, int bufsiz, int item, void* vout)
{
   int count;
   char* ptr;
   char* rparen;

   if (item == PROC_STAT_UID || item == PROC_STAT_PID) {
      /* not really part of stat */
      errno = EIO;
      return PROC_STAT_FAILURE;
   }

   rparen = rindex(line, ')');
   if (rparen == NULL || rparen >= line + bufsiz) {
      errno = EIO;
      return PROC_STAT_FAILURE;
   }

   if (item == PROC_STAT_CMD) {
      char** cout = (char**)vout;
      char* lparen;
      int len;

      lparen = index(line, '(');
      if (lparen == NULL) {
         errno = EIO;
         return PROC_STAT_FAILURE;
      }
      len = (rparen - lparen) - 1;
      *cout = strndup(lparen+1, len);
      if (*cout == NULL) {
         errno = EIO;
         return PROC_STAT_FAILURE;
      }
      return PROC_STAT_SUCCESS;
   }
      
   count = 2;
   ptr = rparen+2;
   while (*ptr && ptr <= line + bufsiz) {
      if (count == item) {
         break;
      } else if (*ptr == ' ') {
        count++;
      }
      ptr++;
   }

   if (ptr >= line + bufsiz) {
      /* past the end so something went wrong */
      errno = EIO;
      return PROC_STAT_FAILURE;
   }

   *(int*)vout = strtoul(ptr, NULL, 10);
   return PROC_STAT_SUCCESS;
}


/*
 *  Sadly, the UID of the process is not one of the stat items
 *  so we need to process this differently.
 */
static int get_proc_uid(pid_t id, int* out)
{
   int ret;
   char buf[PROC_PATH_BUFSIZ];
   struct stat info;

   ret = snprintf(buf, PROC_PATH_BUFSIZ, "/proc/%d", id);
   if (ret <= 0 || ret >= PROC_PATH_BUFSIZ) {
      return PROC_STAT_FAILURE;
   }
   ret = stat(buf, &info);
   if (ret == -1) {
      return PROC_STAT_FAILURE;
   }

   *out = info.st_uid;
   return PROC_STAT_SUCCESS;
}


/*
 *  Sadly, the PID of the process (a.k.a. TGID) is not one of
 *  the stat items so we need to process this differently.
 */
static int get_proc_pid(pid_t id, int* out)
{
   char line[PROC_STATUS_BUFSIZ];
   FILE* status_file;
   pid_t pid = 0;

   status_file = open_proc_status(id);
   if (status_file == NULL) {
      errno = EIO;
      return PROC_STAT_FAILURE;
   }

   while (read_line(status_file, line, PROC_STATUS_BUFSIZ)) {
      if (strncmp(line, "Tgid:\t", 6) == 0) {
         pid = strtol(&line[6], NULL, 10);
         break;
      }
   }
   close_proc_status(status_file);
   *out = pid;
   return PROC_STAT_SUCCESS;
}


/*
 *  Sadly, the name of the process looks truncated in the stat
 *  output.  Use the first entry of /proc/pid/cmdline if set.
 */
static int get_proc_cmd(pid_t id, char** out)
{
   char line[PROC_CMDLINE_BUFSIZ];
   FILE* cmdline_file;
   int ret;
   char* start;
   char* end;

   cmdline_file = open_proc_cmdline(id);
   if (cmdline_file == NULL) {
      errno = EIO;
      return PROC_STAT_FAILURE;
   }

   ret = read_line(cmdline_file, line, PROC_CMDLINE_BUFSIZ);
   if (ret == 0) {
      close_proc_cmdline(cmdline_file);
      return PROC_STAT_FAILURE;
   }

   /*
    *  Some processes modify argv[0] with junk.  Look for the first
    *  space character and nuke it and everything following it.
    */
   end = index(line, ' ');
   if (end && end < line + PROC_CMDLINE_BUFSIZ) {
      *end = 0;
   }

   /*
    *  Remove any leading path.
    */
   start = basename(line);

   *out = strndup(start, PROC_CMDLINE_BUFSIZ);
   if (*out == NULL) {
      close_proc_cmdline(cmdline_file);
      return PROC_STAT_FAILURE;
   }

   close_proc_cmdline(cmdline_file);
   return PROC_STAT_SUCCESS;
}


/***********************************************************************
 *                                                                     *
 *  Miscellaneous utility functions                                    *
 *                                                                     *
 ***********************************************************************/


/*
 *  Fill buf with next line from file, return line length.
 */
static int read_line(FILE* file, char* buf, int bufsiz)
{
   int count = 0;
   int c;

   while ((c = getc(file)) != EOF && count < bufsiz) {
      buf[count++] = c;
      if (c == '\n') {
         buf[count] = '\0';
         break;
      }
   }

   return count;
}


/***********************************************************************
 *                                                                     *
 *  External entry points                                              *
 *                                                                     *
 ***********************************************************************/

int proc_stat(pid_t id, int item, void* vout)
{
   int ret;
   FILE* stat_file;
   char buf[PROC_STAT_BUFSIZ];
   int* out = (int*)vout;

   if (item == PROC_STAT_UID) {
      return get_proc_uid(id, out);
   }

   if (item == PROC_STAT_PID) {
      return get_proc_pid(id, out);
   }

   stat_file = open_proc_stat(id);
   if (stat_file == NULL) {
      errno = EIO;
      return PROC_STAT_FAILURE;
   }

   ret = read_line(stat_file, buf, PROC_STAT_BUFSIZ);
   if (ret == 0) {
      errno = EIO;
      close_proc_stat(stat_file);
      return PROC_STAT_FAILURE;
   }

   ret = extract_stat(buf, PROC_STAT_BUFSIZ, item, out);
   if (ret == PROC_STAT_FAILURE) {
      errno = EIO;
      close_proc_stat(stat_file);
      return PROC_STAT_FAILURE;
   }

   close_proc_stat(stat_file);

   if (item == PROC_STAT_CMD) {
      char** cout = (char**)vout;
      char* cmdline_name;
      int len;

      len = strlen(*cout);
      if (len == 15) {
         /*
          *  The name in stat has been potentially truncated.
          *  See if we can get a better name from the first slot in
          *  /proc/pid/cmdline.  Otherwise, just use the stat name.
          */
         ret = get_proc_cmd(id, &cmdline_name);
         if (ret == PROC_STAT_SUCCESS) {
            free(*cout);
            *cout = cmdline_name;
         }
      }
   }

   return PROC_STAT_SUCCESS;
}


/*
 *  Silenetly alias pid_stat to proc_stat for backward compatibility
 *  with old dynamically linked binaries.
 */
int pid_stat(pid_t, int, void*);
int pid_stat(pid_t id, int item, void* vout) {
   return proc_stat(id, item, vout);
}
