/*
 *  Copyright (C) 2001  Concurrent Computer Corporation
 *  This source is covered by the GNU LGPL Version 2.1
 *  Author: Jason Baietto (jason.baietto@ccur.com)
 *
 *  Functions to walk the /proc process tree.
 *
 *  Updated to know about /proc/.tid directories.
 *  Updated to walk 2.6 /proc/pid/task directories.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <dirent.h>
#include <string.h>
#include "proc_stat.h"
#include "proc_walk.h"

#define PROC_PATH_BUFSIZ 1024

#define TID_MODE_UNINITIALIZED      0
#define TID_MODE_INDISTINGUISHABLE  1
#define TID_MODE_HIERARCHICAL       2
#define TID_MODE_DOT_TID            3

#define MAX_D_NAME_LEN 256


/*==========================================================================*
 *                                                                          *
 *  Private data and functions.                                             *
 *                                                                          *
 *==========================================================================*/


/*
 *  Note that the pid info is valid during the entire walk, but the
 *  tid info is only valid when get_next_tid() is being used.
 *  The indexes always point to the next entry after a get call.
 */
struct proc_walk_id {
   /*
    *  Fields for walking processes.
    */
   struct dirent** pid_name_array;
   int pid_name_index;
   int pid_name_count;
   /*
    *  Fields for walking threads.
    */
   struct dirent** tid_name_array;
   int tid_name_index;
   int tid_name_count;
   int tid_mode;
   /*
    *  The following fields are only used on systems with /proc/.tid
    */
   struct dirent** all_tids_name_array;
   pid_t* all_tids_pid_array;
   int all_tids_count;
};


static void init_pwid(pwid_t* pwid) {
   pwid->pid_name_array = NULL;
   pwid->pid_name_index = 0;
   pwid->pid_name_count = 0;
   pwid->tid_name_array = NULL;
   pwid->tid_name_index = 0;
   pwid->tid_name_count = 0;
   pwid->tid_mode = TID_MODE_UNINITIALIZED;
   pwid->all_tids_name_array = NULL;
   pwid->all_tids_pid_array = NULL;
   pwid->all_tids_count = 0;
}


static int proc_select(const struct dirent* dir)
{
   return isdigit(dir->d_name[0]);
}


static int thread_select(const struct dirent* dir)
{
   return (dir->d_name[0] == '.' && isdigit(dir->d_name[1]));
}


static int proc_compare(const void* a, const void* b)
{
   int a_pid;
   int b_pid;
   const struct dirent* a_dir = *(const struct dirent * const *)a;
   const struct dirent* b_dir = *(const struct dirent * const *)b;
   a_pid = strtol(a_dir->d_name, NULL, 10);
   b_pid = strtol(b_dir->d_name, NULL, 10);
   return (a_pid - b_pid);
}


static void free_pid_info(pwid_t* pwid) {
   int i;

   if (pwid->pid_name_array == NULL) {
      return;
   }
   for (i=0; i < pwid->pid_name_count; i++) {
      free(pwid->pid_name_array[i]);
   }
   free(pwid->pid_name_array);
   pwid->pid_name_array = NULL;
   pwid->pid_name_count = 0;
   pwid->pid_name_index = 0;
}


static void free_tid_info(pwid_t* pwid) {
   int i;

   if (pwid->tid_name_array == NULL) {
      return;
   }
   for (i=0; i < pwid->tid_name_count; i++) {
      free(pwid->tid_name_array[i]);
   }
   free(pwid->tid_name_array);
   pwid->tid_name_array = NULL;
   pwid->tid_name_count = 0;
   pwid->tid_name_index = 0;
}


static void free_all_tids_info(pwid_t* pwid) {
   int i;

   if (pwid->all_tids_name_array == NULL) {
      return;
   }
   for (i=0; i < pwid->all_tids_count; i++) {
      free(pwid->all_tids_name_array[i]);
   }
   free(pwid->all_tids_name_array);
   if (pwid->all_tids_pid_array) {
      free(pwid->all_tids_pid_array);
   }
   pwid->all_tids_name_array = NULL;
   pwid->all_tids_pid_array = NULL;
   pwid->all_tids_count = 0;
}


/*
 *  Routine to abstract the different representations of threads
 *  among various /proc implementations.  This is optimized for the
 *  case where task sub-directories exist, which is the norm for
 *  most recent kernels.  For other cases, task-like dirents are faked.
 *
 *  Returns number of tids loaded, or 0 for unrecoverable errors.
 */
static int load_tids_for_pid(pwid_t* pwid, pid_t pid) {
   char buffer[PROC_PATH_BUFSIZ];
   int ret;

   if (pwid->tid_mode == TID_MODE_UNINITIALIZED) {
      struct stat st;
      ret = stat("/proc/1/task", &st);
      if (ret == 0) {
         /*
          *  This is the simple and clean mode (thanks Albert!).
          */
         pwid->tid_mode = TID_MODE_HIERARCHICAL;
      } else {
         /*
          *  There's no easy way to map tids to pids in these modes.
          *  Look for and process any /proc/.tid directories.
          */
         ret = scandir("/proc", &pwid->all_tids_name_array, thread_select, proc_compare);
         if (ret <= 0) {
            /*
             *  Either there are no existing threads or this /proc does not
             *  distinguish between threads and pids (various 2.4 kernels).
             *  The pid will be faked up to look like the sole tid of the pid.
             *  A get_next_tid walk will be identical to a get_next_pid walk.
             */
            pwid->tid_mode = TID_MODE_INDISTINGUISHABLE;
         } else {
            /*
             *  Found some /proc/.tid dirs, so build and cache a tid-to-pid
             *  map up-front for use during the entire walk.
             */
            int i;
            pwid->tid_mode = TID_MODE_DOT_TID;
            pwid->all_tids_count = ret;
            pwid->all_tids_pid_array = (pid_t*)malloc(ret * sizeof(pid_t));
            if (pwid->all_tids_pid_array == NULL) {
               return 0;
            }
            for (i=0; i < pwid->all_tids_count; i++) {
               pid_t owner = 0;
               pid_t tid;
               tid = strtol(&pwid->all_tids_name_array[i]->d_name[1], NULL, 10);
               ret = proc_stat(tid, PROC_STAT_PID, &owner);
               if (ret == PROC_STAT_SUCCESS) {
                  pwid->all_tids_pid_array[i] = owner;
               } else {
                  pwid->all_tids_pid_array[i] = 0;
               }
            }
         }
      }
   }

   switch (pwid->tid_mode) {

      case TID_MODE_HIERARCHICAL: {
         /*
          *  Read in the directory entries for the task directory.
          */
         ret = snprintf(buffer, PROC_PATH_BUFSIZ, "/proc/%d/task", pid);
         if (ret <= 0 || ret >= PROC_PATH_BUFSIZ) {
            return 0;
         }
         ret = scandir(buffer, &pwid->tid_name_array, proc_select, proc_compare);
         if (ret > 0) {
            pwid->tid_name_index = 0;
            pwid->tid_name_count = ret;
         } else {
            /*
             *  Oops, the process looks like it already exited.  Fake up a
             *  tid matching the original pid for historical reasons, and
             *  to allow the walk to continue smoothly.
             */
            pwid->tid_name_index = 0;
            pwid->tid_name_count = 1;
            pwid->tid_name_array = (struct dirent**)malloc(sizeof(struct dirent*));
            if (pwid->tid_name_array == NULL) {
               return 0;
            }

            pwid->tid_name_array[0] = (struct dirent*)malloc(sizeof(struct dirent));
            if (pwid->tid_name_array[0] == NULL) {
               return 0;
            }
            memset(pwid->tid_name_array[0], 0, sizeof(struct dirent));

            ret = snprintf(pwid->tid_name_array[0]->d_name, MAX_D_NAME_LEN, "%d", pid);
            if (ret <= 0 || ret >= MAX_D_NAME_LEN) {
               return 0;
            }
         }
         break;
      }

      case TID_MODE_DOT_TID: {
         /*
          *  Look for tids with a matching pid and fake up dirent entries.
          *  The first faked entry is always the pid itself.
          */
         int matches;
         int entry;
         int i;

         /*
          *  Count the matching tids (plus 1 for the pid itself).
          */
         matches = 1;
         for (i=0; i < pwid->all_tids_count; i++) {
            if (pwid->all_tids_pid_array[i] == pid) {
               matches++;
            }
         }

         pwid->tid_name_index = 0;
         pwid->tid_name_count = matches;
         pwid->tid_name_array = (struct dirent**)malloc(matches * sizeof(struct dirent*));
         if (pwid->tid_name_array == NULL) {
            return 0;
         }
         memset(pwid->tid_name_array, 0, matches * sizeof(struct dirent*));

         /*
          *  Fill in the index array with pointers to newly allocated dirents.
          */
         for (i=0; i < matches; i++) {
            pwid->tid_name_array[i] = (struct dirent*)malloc(sizeof(struct dirent));
            if (pwid->tid_name_array[i] == NULL) {
               return 0;
            }
            memset(pwid->tid_name_array[i], 0, sizeof(struct dirent));
         }

         /*
          *  Fake the first entry to match the pid.
          */
         ret = snprintf(pwid->tid_name_array[0]->d_name, MAX_D_NAME_LEN, "%d", pid);
         if (ret <= 0 || ret >= MAX_D_NAME_LEN) {
            return 0;
         }

         /*
          *  Fake the remaining matching entries.
          */         
         entry = 1;
         for (i=0; i < pwid->all_tids_count; i++) {
            if (pwid->all_tids_pid_array[i] == pid) {
               strncpy(pwid->tid_name_array[entry]->d_name, &pwid->all_tids_name_array[i]->d_name[1], MAX_D_NAME_LEN);
               if (pwid->tid_name_array[entry]->d_name[MAX_D_NAME_LEN-1] != 0) {
                  return 0;
               }
               entry++;
            }
         }

         break;
      }

      case TID_MODE_INDISTINGUISHABLE: {
         /*
          *  Pretend that every pid has a single matching tid.
          */

         pwid->tid_name_index = 0;
         pwid->tid_name_count = 1;
         pwid->tid_name_array = (struct dirent**)malloc(sizeof(struct dirent*));
         if (pwid->tid_name_array == NULL) {
            return 0;
         }

         pwid->tid_name_array[0] = (struct dirent*)malloc(sizeof(struct dirent));
         if (pwid->tid_name_array[0] == NULL) {
            return 0;
         }
         memset(pwid->tid_name_array[0], 0, sizeof(struct dirent));

         ret = snprintf(pwid->tid_name_array[0]->d_name, MAX_D_NAME_LEN, "%d", pid);
         if (ret <= 0 || ret >= MAX_D_NAME_LEN) {
            return 0;
         }
         break;
      }

      default: {
         return 0;
      }
   }

   return pwid->tid_name_count;
}


/*==========================================================================*
 *                                                                          *
 *  Public entry points.                                                    *
 *                                                                          *
 *==========================================================================*/


/*
 *  Return pwid pointer, or NULL for unrecoverable errors.
 */
pwid_t* proc_walk_init(void)
{
   int ret;
   pwid_t* pwid;

   pwid = (pwid_t*)malloc(sizeof(pwid_t));
   if (pwid == NULL) {
      return NULL;
   }

   init_pwid(pwid);

   ret = scandir("/proc", &pwid->pid_name_array, proc_select, proc_compare);
   if (ret <= 0) {
      free(pwid);
      return NULL;
   }
   pwid->pid_name_count = ret;

   return pwid;
}


/*
 *  Return next pid in list, or 0 for done or unrecoverable errors.
 */
int proc_walk_get_next_pid(pwid_t* pwid)
{
   int pid;
   
   if (pwid == NULL || pwid->pid_name_array == NULL) {
      return 0;
   }

   if (pwid->pid_name_index >= pwid->pid_name_count) {
      return 0;
   }

   free_tid_info(pwid);

   pid = strtol(pwid->pid_name_array[pwid->pid_name_index]->d_name, NULL, 10);

   pwid->pid_name_index++;
   return pid;
}


/*
 *  Return the current pid in the list, or 0 for done or unrecoverable errors.
 *  This is also useful to find the pid that goes with the current tid.
 */
int proc_walk_get_pid(pwid_t* pwid) {
   int pid;
   int i;

   if (pwid == NULL || pwid->pid_name_array == NULL) {
      return 0;
   }

   i = pwid->pid_name_index - 1;
   if (i < 0) {
      /*
       *  Handle being called before get_next_pid.
       */
      return proc_walk_get_next_pid(pwid);
   }

   if (i >= pwid->pid_name_count) {
      return 0;
   }

   pid = strtol(pwid->pid_name_array[i]->d_name, NULL, 10);

   return pid;
}


/*
 *  Return next tid in list, 0 for done or unrecoverable errors.
 *  Happily moves on to next pid if tids for current pid are exhausted,
 *  so this can be used to easily walk through all threads in the system.
 */
int proc_walk_get_next_tid(pwid_t* pwid)
{
   int ret;
   int pid;
   int tid;

   if (pwid == NULL || pwid->pid_name_array == NULL) {
      return 0;
   }

   if (pwid->tid_name_array && pwid->tid_name_index >= pwid->tid_name_count) {
      /*
       *  Reached end of tids for the current pid.
       */
      pid = proc_walk_get_next_pid(pwid);
      if (pid == 0) {
         return 0;
      }
   }

   if (pwid->tid_name_array == NULL) {
      /*
       *  First time tids for the current pid are being examined.
       */
      pid = proc_walk_get_pid(pwid);
      if (pid == 0) {
         return 0;
      }
      ret = load_tids_for_pid(pwid, pid);
      if (ret == 0) {
         return 0;
      }
   }

   /*
    *  Note that there will always be at least one tid for every pid.
    */
   tid = strtol(pwid->tid_name_array[pwid->tid_name_index]->d_name, NULL, 10);

   pwid->tid_name_index++;
   return tid;
}


/*
 *  Reset the list back to the first element so it can be
 *  traversed from the start again.
 */
void proc_walk_reset(pwid_t* pwid)
{
   pwid->pid_name_index = 0;
   free_tid_info(pwid);
}


/*
 *  Clean up all allocated resources.
 */
void proc_walk_done(pwid_t* pwid)
{
   if (pwid == NULL) {
      return;
   }
   free_pid_info(pwid);
   free_tid_info(pwid);
   free_all_tids_info(pwid);
   free(pwid);
}


/*==========================================================================*
 *                                                                          *
 *  Thread walk support for just walking a single process's threads.        *
 *                                                                          *
 *==========================================================================*/


/*
 *  The thread_walk_id type is identical to the proc_walk_id type and is
 *  only given a different name to discourage the unsupported intermixing
 *  of thread_walk and proc_walk calls.
 */
struct thread_walk_id {
   /*
    *  Fields for walking processes.
    */
   struct dirent** pid_name_array;
   int pid_name_index;
   int pid_name_count;
   /*
    *  Fields for walking threads.
    */
   struct dirent** tid_name_array;
   int tid_name_index;
   int tid_name_count;
   int tid_mode;
   /*
    *  The following fields are only used on systems with /proc/.tid
    */
   struct dirent** all_tids_name_array;
   pid_t* all_tids_pid_array;
   int all_tids_count;
};


/*
 *  Returns twid pointer, or NULL for unrecoverable errors.
 */
twid_t* thread_walk_init(pid_t id)
{
   int ret;
   pid_t pid;
   twid_t* twid;

   /*
    *  Ensure that the specified pid exists, and also translate any
    *  tid specified into its containing pid (allows sibling walks).
    */
   ret = proc_stat(id, PROC_STAT_PID, &pid);
   if (ret == PROC_STAT_FAILURE) {
      return NULL;
   }

   twid = (twid_t*)malloc(sizeof(twid_t));
   if (twid == NULL) {
      return NULL;
   }

   init_pwid((pwid_t*)twid);

   ret = load_tids_for_pid((pwid_t*)twid, pid);
   if (ret == 0) {
      free(twid);
      return NULL;
   }

   return twid;
}


/*
 *  Return next tid in list, 0 for done or unrecoverable errors.
 */
int thread_walk_get_next_tid(twid_t* twid)
{
   int tid;
   
   if (twid == NULL || twid->tid_name_array == NULL) {
      return 0;
   }

   if (twid->tid_name_index >= twid->tid_name_count) {
      return 0;
   }

   tid = strtol(twid->tid_name_array[twid->tid_name_index]->d_name, NULL, 10);

   twid->tid_name_index++;
   return tid;
}


void thread_walk_reset(twid_t* twid)
{
   twid->tid_name_index = 0;
}


void thread_walk_done(twid_t* twid)
{
   if (twid == NULL) {
      return;
   }
   free_tid_info((pwid_t*)twid);
   free_all_tids_info((pwid_t*)twid);
   free(twid);
}
