/*
 *  Copyright (C) 2003  Concurrent Computer Corporation
 *  This source is covered by the GNU GPL Version 2.
 *  Author: Jason Baietto (jason.baietto@ccur.com)
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <getopt.h>
#include <errno.h>
#include <libgen.h>
#include <stdarg.h>
#include <string.h>
#include <sched.h>
#include <pwd.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <mpadvise.h>
#include <proc_walk.h>
#include <proc_stat.h>
#include <cpuset.h>
#include <asm/unistd.h>

#ifndef VERSION_MAJOR
#define VERSION_MAJOR 3
#endif
#ifndef VERSION_MINOR
#define VERSION_MINOR 0
#endif
#ifndef VERSION_TINY
#define VERSION_TINY 0
#endif

#define RUN_BUFSIZ    1024
#define MAX_COPIES    1024

#define OPT_UNKNOWN     '?'
#define OPT_ARG_MISSING ':'
#define OPT_BIAS        'b'
#define OPT_POLICY      's'
#define OPT_PRIORITY    'P'
#define OPT_ALL         'a'
#define OPT_PID         'p'
#define OPT_TID         't'
#define OPT_LWPID       'l'
#define OPT_GROUP       'g'
#define OPT_USER        'u'
#define OPT_NAME        'n'
#define OPT_NEGATE      'N'
#define OPT_QUANTUM     'q'
#define OPT_COPIES      'c'
#define OPT_FORK        'f'
#define OPT_INFO        'i'
#define OPT_VERBOSE     'v'
#define OPT_VERSION     'V'
#define OPT_HELP        'h'
#define OPT_CPUS        'C'
#define OPT_DEBUG       'd'

/*
 *  The '+' forces option processing to stop as soon as a
 *  non-option argument is encountered.
 */
static const char* short_options = "+b:s:P:p:t:l:g:u:n:q:c:fiNvVhCda";

static const struct option long_options[] = {
   { "bias",        required_argument, 0, OPT_BIAS },
   { "policy",      required_argument, 0, OPT_POLICY },
   { "priority",    required_argument, 0, OPT_PRIORITY },
   { "all",         no_argument,       0, OPT_ALL },
   { "pid",         required_argument, 0, OPT_PID },
   { "tid",         required_argument, 0, OPT_TID },
   { "lwpid",       required_argument, 0, OPT_LWPID },
   { "group",       required_argument, 0, OPT_GROUP },
   { "user",        required_argument, 0, OPT_USER },
   { "name",        required_argument, 0, OPT_NAME },
   { "quantum",     required_argument, 0, OPT_QUANTUM },
   { "copies",      required_argument, 0, OPT_COPIES },
   { "negate",      no_argument,       0, OPT_NEGATE },
   { "fork",        no_argument,       0, OPT_FORK },
   { "info",        no_argument,       0, OPT_INFO },
   { "verbose",     no_argument,       0, OPT_VERBOSE },
   { "version",     no_argument,       0, OPT_VERSION },
   { "help",        no_argument,       0, OPT_HELP },
   { "cpus",        no_argument,       0, OPT_CPUS },
   { "debug",       no_argument,       0, OPT_DEBUG },
   { 0, 0, 0, 0 }
};

static const char help[] = 
"Usage: %s [OPTIONS] { COMMAND [ARGS] | PROCESS/THREAD_SPECIFIER }\n"
"Set scheduling parameters and CPU bias for a new process or a\n"
"list of existing processes/threads.\n"
"\n"
"OPTIONS can be one or more of the following options:\n"
"\n"
"   -b, --bias=LIST        Set the CPU bias to the LIST of CPUs;\n"
"                          CPUs are numbered starting from 0\n"
"   -s, --policy=POLICY    Set the scheduling policy to POLICY\n"
"                          (SCHED_OTHER, SCHED_RR or SCHED_FIFO)\n"
"   -P, --priority=LEVEL   Set the scheduling priority to LEVEL;\n"
"                          SCHED_FIFO and SCHED_RR range: 1 to 99\n"
"                          SCHED_OTHER: only priority 0 is valid\n"
"   -q, --quantum=QUANTUM  Set the SCHED_RR quantum to QUANTUM;\n"
"                          use --quantum=list for valid settings\n"
"   -N, --negate           Negate the CPU bias list; all CPUs\n"
"                          except those listed will be selected\n"
"   -f, --fork             Fork COMMAND and return immediately\n"
"   -c, --copies=COUNT     Run COUNT identical copies of COMMAND\n"
"   -i, --info             Output process environment information\n"
"   -V, --version          Output version information and exit\n"
"   -v, --verbose          Output information before each action\n"
"   -h, --help             Display this help and exit\n"
"\n"
"PROCESS/THREAD_SPECIFIER is exactly one of the following options:\n"
"\n"
"   -a, --all              Select all existing processes and threads\n"
"   -p, --pid=LIST         Specify LIST of existing PIDs to modify\n"
"   -t, --tid=LIST         Specify LIST of existing TIDs to modify\n"
"   -g, --group=LIST       Specify LIST of process groups to modify; all\n"
"                          existing processes in the groups will be modified\n"
"   -u, --user=LIST        Specify LIST of users to modify; all existing\n"
"                          processes owned by the users will be modified\n"
"   -n, --name=LIST        Specify LIST of existing process names to modify\n"
"\n"
"Multiple comma separated values can be specified for all LISTs and ranges\n"
"are allowed where appropriate (e.g. \"run -b 0,2-5 autopilot\").\n"
"\n"
"See the run(1) man page for more information.\n";


static const char version[] = 
#if (VERSION_TINY == 0)
"%s %d.%d\n"
#else
"%s %d.%d.%d\n"
#endif
"Written by Jason Baietto <jason.baietto@ccur.com>\n"
"\n"
"Copyright (C) 2001 Concurrent Computer Corporation\n"
"This is free software; see the source for copying conditions.  There is NO\n"
"warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n";

static const char *prog_name;
static int opt_bias = 0;
static int opt_policy = 0;
static int opt_priority = 0;
static int opt_quantum = 0;
static int opt_all = 0;
static int opt_pid = 0;
static int opt_tid = 0;
static int opt_group = 0;
static int opt_user = 0;
static int opt_name = 0;
static int opt_verbose = 0;
static int opt_fork = 0;
static int opt_negate = 0;
static int opt_copies = 0;
static int opt_info = 0;
static int opt_debug = 0;

static int policy;
static int priority;
static int quantum_arg;
static int command;
static int copies = 1;

static cpuset_t* bias_cpu_setp;
static cpuset_t* valid_cpu_setp;

typedef struct int_list {
   int data;
   struct int_list* next;
} int_list_t;

typedef struct string_list {
   char* data;
   struct string_list* next;
} string_list_t;

static int_list_t* tid_list;
static int_list_t* pid_list;
static int_list_t* uid_list;
static int_list_t* pgrp_list;
static string_list_t* name_list;

static int error_count;


/***********************************************************************
 *                                                                     *
 *  Miscellaneous utility functions                                    *
 *                                                                     *
 ***********************************************************************/


/*
 *  Provide a local gettid entry point.  On systems that do not support
 *  gettid only process level functionality will be provided.
 */
static pid_t gettid(void) {
   pid_t tid;
#ifdef __NR_gettid
   tid = (pid_t)syscall(__NR_gettid);
   if (errno == ENOSYS)
      /* fall through */
#endif
      tid = getpid();
   return tid;
}


/*
 *  Keep the program name dynamic in case we want to create
 *  a rerun link to placate existing PowerMAX users.
 */
static void set_prog_name(char* path)
{
   prog_name = basename(path);
   if (prog_name[0] == '\0') {
      prog_name = "run";
   }
}


static void print_help(void)
{
   printf(help, prog_name);
   exit(0);
}


static void print_version(void)
{
#if (VERSION_TINY == 0)
   printf(version, prog_name, VERSION_MAJOR, VERSION_MINOR);
#else
   printf(version, prog_name, VERSION_MAJOR, VERSION_MINOR, VERSION_TINY);
#endif
   exit(0);
}

#ifdef __GNUC__
/* make gcc do printf-style type checking on variable args */
static void info (const char* format, ...) __attribute__ ((format (printf, 1, 2)));
static void warn (const char* format, ...) __attribute__ ((format (printf, 1, 2)));
static void fatal(const char* format, ...) __attribute__ ((format (printf, 1, 2)));
static void debug(const char* format, ...) __attribute__ ((format (printf, 1, 2)));
#endif

static void info(const char* format, ...)
{
   va_list ap;
   char buf[RUN_BUFSIZ];

   snprintf(buf, RUN_BUFSIZ, "%s: %s\n", prog_name, format);
   va_start(ap, format);
   vfprintf(stderr, buf, ap);
   va_end(ap);
}


static void warn(const char* format, ...)
{
   va_list ap;
   char buf[RUN_BUFSIZ];

   snprintf(buf, RUN_BUFSIZ, "%s: %s\n", prog_name, format);
   va_start(ap, format);
   vfprintf(stderr, buf, ap);
   va_end(ap);
   error_count++;
}


static void fatal(const char* format, ...)
{
   va_list ap;
   char buf[RUN_BUFSIZ];

   if (opt_debug && errno) {
      perror(prog_name);
   }
   snprintf(buf, RUN_BUFSIZ, "%s: %s\n", prog_name, format);
   va_start(ap, format);
   vfprintf(stderr, buf, ap);
   va_end(ap);
   exit(1);
}


static void internal_error(void)
{
   fatal("internal error");
}


static void debug(const char* format, ...)
{
   va_list ap;
   char buf[RUN_BUFSIZ];

   if (opt_debug) {
      snprintf(buf, RUN_BUFSIZ, "%s: DEBUG: %s\n", prog_name, format);
      va_start(ap, format);
      vfprintf(stderr, buf, ap);
      va_end(ap);
   }
}


static void init_valid_cpu_set(void) {
   int ret;

   if (valid_cpu_setp) {
      return;
   }

   valid_cpu_setp = cpuset_alloc();
   if (!valid_cpu_setp) {
      fatal("out of memory");
   }

   ret = mpadvise(MPA_CPU_ACTIVE, 0, 0, valid_cpu_setp);
   if (ret == MPA_FAILURE) {
      fatal("unable to access cpu info");
   }

   if (cpuset_is_empty(valid_cpu_setp)) {
      fatal("no active cpus found!");
   }
}


static void validate_cpu(int cpu)
{
   init_valid_cpu_set();

   if (!cpuset_is_valid_cpu(cpu) || !cpuset_get_cpu(valid_cpu_setp, cpu)) {
      fatal("invalid cpu %d", cpu);
   }
}


static int is_only_digits(const char* str)
{
   if (!str) {
      return 0;
   }
   while (*str) {
      if (!isdigit(*str)) {
         return 0;
      }
      str++;
   }
   return 1;
}


static int is_ccur_kernel(void) {
   struct stat st;
   int ret = stat("/proc/ccur", &st);
   if (ret == 0) {
      debug("running ccur kernel");
      return 1;
   } else {
      debug("running non-ccur kernel");
      return 0;
   }
}


/***********************************************************************
 *                                                                     *
 *  Simple integer linked-list routines                                *
 *                                                                     *
 ***********************************************************************/


static int_list_t* alloc_int_node(int data)
{
   int_list_t* tmp;

   tmp = (int_list_t*)malloc(sizeof(int_list_t));
   if (tmp == NULL) {
      fatal("malloc failed");
   }
   tmp->data = data;
   return tmp;
}


static void free_int_node(int_list_t* node)
{
   if (node) {
      free(node);
   }
}


static void insert_int(int_list_t** list, int data)
{
   int_list_t* node;

   if (!list) {
      internal_error();
   }

   node = alloc_int_node(data);
   node->next = *list;
   *list = node;
}


static void free_int_list(int_list_t** list)
{
   int_list_t* node = *list;
   int_list_t* next;

   while (node) {
      next = node->next;
      free_int_node(node);
      node = next;
   }
   *list = NULL;
}


/***********************************************************************
 *                                                                     *
 *  Simple string linked-list routines                                 *
 *                                                                     *
 ***********************************************************************/


static string_list_t* alloc_string_node(const char* data)
{
   string_list_t* tmp;

   tmp = (string_list_t*)malloc(sizeof(string_list_t));
   if (tmp == NULL) {
      fatal("malloc failed");
   }
   tmp->data = strdup(data);
   if (tmp->data == NULL) {
      fatal("malloc failed");
   }
   return tmp;
}


static void free_string_node(string_list_t* node)
{
   if (node) {   
      if (node->data) {
         free(node->data);
      }
      free(node);
   }
}


static void insert_string(string_list_t** list, char* data)
{
   string_list_t* node;

   if (!list) {
      internal_error();
   }

   node = alloc_string_node(data);
   node->next = *list;
   *list = node;
}


static void free_string_list(string_list_t** list)
{
   string_list_t* node = *list;
   string_list_t* next;

   while (node) {
      next = node->next;
      free_string_node(node);
      node = next;
   }
   *list = NULL;
}


/***********************************************************************
 *                                                                     *
 *  List parsing routines                                              *
 *                                                                     *
 ***********************************************************************/


/*
 *  Given an integer list string, return an int_list.
 *  Comma separated values and ranges are allowed.
 *  All numbers must be positive base 10 integers.
 *  This is not very efficient for large ranges.
 */
static int_list_t* parse_int_list(const char* arg, const char* msg)
{
   int value;
   int last_value = -1;
   int separator;
   int count;
   int ret;
   int_list_t* list = NULL;



   while (1) {
      /*
       *  Scan a number.
       */
      ret = sscanf(arg, "%d%n", &value, &count);
      if (ret != 1) {
         fatal("invalid %s syntax (number expected)", msg);
      }

      if (value < 0) {
         fatal("invalid %s syntax (positive number expected)", msg);
      }

      if (last_value == -1) {
         insert_int(&list, value);
      } else {
         int i;
         for (i = last_value; i <= value; i++) {
            insert_int(&list, i);
         }
         last_value = -1;
      }

      /*
       *  Check for eol.
       */
      arg += count;
      if (arg[0] == '\0') {
         return list;
      }

      /*
       *  Scan a separator.
       */
      separator = *arg;
      arg++;
      if (separator == '-') {
         last_value = value;
      } else if (separator != ',') {
         fatal("invalid %s syntax (separator expected)", msg);
         exit(1);
      }
   }
}


/*
 *  Given a string list string, return an string_list.
 *  Strings may not contain commas.  Ranges are not supported.
 */
static string_list_t* parse_string_list(const char* arg) {
   char buf[RUN_BUFSIZ];
   char* ptr = buf;
   string_list_t* list = NULL;

   while (*arg != 0) {
      if (*arg == ',') {
         *ptr = 0;
         insert_string(&list, buf);
         ptr = buf;
         arg++;
      } else {
         *ptr++ = *arg++;
      }
   }
   if (ptr != buf) {
      *ptr = 0;
      insert_string(&list, buf);
   }

   return list;
}


/***********************************************************************
 *                                                                     *
 *  Quantum manipulation functions                                     *
 *                                                                     *
 ***********************************************************************/

typedef struct quantum_entry {
   int nice;
   int msec;
} quantum_entry_t;

/*
 *  This table is now dynamically generated based on /proc/sys/sched files,
 *  however for backward compatibility we still use these default values
 *  on systems without /proc/sys/sched.
 */
static quantum_entry_t quantum_map[] = {
   {-20, 199},
   {-19, 194},
   {-18, 189},
   {-17, 184},
   {-16, 184},
   {-15, 174},
   {-14, 169},
   {-13, 164},
   {-12, 164},
   {-11, 155},
   {-10, 150},
   {-9,  145},
   {-8,  145},
   {-7,  135},
   {-6,  130},
   {-5,  125},
   {-4,  125},
   {-3,  116},
   {-2,  111},
   {-1,  106},
   {0,   106},
   {1,   96},
   {2,   91},
   {3,   86},
   {4,   86},
   {5,   77},
   {6,   72},
   {7,   67},
   {8,   67},
   {9,   57},
   {10,  52},
   {11,  47},
   {12,  47},
   {13,  38},
   {14,  33},
   {15,  28},
   {16,  28},
   {17,  18},
   {18,  13},
   {19,  13},
};

#define QUANTUM_FAILURE 0
#define QUANTUM_SUCCESS 1
#define MIN_TIMESLICE_FILE "/proc/sys/sched/min_timeslice"
#define MAX_TIMESLICE_FILE "/proc/sys/sched/max_timeslice"
#define MIN_NICE -20
#define MAX_NICE 19
#define NICE_VALUES (MAX_NICE-MIN_NICE)

/*
 *  Dynamically determine the available range of valid time quanta.
 *  Note that any premature return keeps the default values in the map
 *  (which were accurate circa RedHawk 1.2).
 */
static void init_quantum_map(void) {
   static int initialized = 0;
   FILE* min_timeslice_file;
   FILE* max_timeslice_file;
   int min_timeslice, max_timeslice;
   unsigned int i;

   if (initialized) {
      return;
   }

   initialized++;

   min_timeslice_file = fopen(MIN_TIMESLICE_FILE, "r");
   if (min_timeslice_file == NULL) {
      debug("assuming system does not support /proc/sys/sched");
      return;
   } else {
      char buf[RUN_BUFSIZ];
      int ret;
      ret = fread(buf, sizeof(char), RUN_BUFSIZ, min_timeslice_file);
      if (ret == 0 || ret >= RUN_BUFSIZ) {
         warn("error reading " MIN_TIMESLICE_FILE " file");
         fclose(min_timeslice_file);
         return;
      }
      min_timeslice = strtoul(buf, NULL, 10);
      if (min_timeslice == 0) {
         warn("invalid value in " MIN_TIMESLICE_FILE " file");
         fclose(min_timeslice_file);
         return;
      }
      fclose(min_timeslice_file);
   }
   
   max_timeslice_file = fopen(MAX_TIMESLICE_FILE, "r");
   if (max_timeslice_file == NULL) {
      debug("assuming system does not support /proc/sys/sched");
      return;
   } else {
      char buf[RUN_BUFSIZ];
      int ret;
      ret = fread(buf, sizeof(char), RUN_BUFSIZ, max_timeslice_file);
      if (ret == 0 || ret >= RUN_BUFSIZ) {
         warn("error reading " MAX_TIMESLICE_FILE " file");
         fclose(max_timeslice_file);
         return;
      }
      max_timeslice = strtoul(buf, NULL, 10);
      if (max_timeslice == 0) {
         warn("invalid value in " MAX_TIMESLICE_FILE " file");
         fclose(max_timeslice_file);
         return;
      }
      fclose(max_timeslice_file);
   }

   debug("new quantum map (min=%d, max=%d)", min_timeslice, max_timeslice);

   for (i=0; i < sizeof(quantum_map)/sizeof(quantum_entry_t); i++) {
      /* translate -19..20 into 39..0 (higher number is higher prio) */
      int prio = NICE_VALUES - (quantum_map[i].nice + MAX_NICE + 1);
      int jiffs = min_timeslice + 
          (((max_timeslice - min_timeslice) * prio)/NICE_VALUES);
      int msec = jiffs * (1000 / HZ);
      quantum_map[i].msec = msec;
   }
}

/*
 *  Given a msec value find the corresponding nice value.
 *  Only exact matches are allowed.
 */
static int quantum_lookup_msec(int msec, int* nice_ptr) {
   unsigned int i;
   unsigned int count = sizeof(quantum_map)/sizeof(quantum_entry_t);

   for (i=0; i < count; i++) {
      if (quantum_map[i].msec == msec) {
         *nice_ptr = quantum_map[i].nice;
         return QUANTUM_SUCCESS;
      }
   }
   return QUANTUM_FAILURE;
}

/*
 *  Given a quantum argument, determine if it has been specified
 *  in msec format and if so return the msec value.
 */
static int quantum_check_msec(const char* arg, int* quantum_ptr) {
   int ret;
   int value = 0;
   char suffix[10] = {0,0,0,0,0,0,0,0,0,0};

   init_quantum_map();

   ret = sscanf(arg, "%d%2c", &value, suffix);
   if (ret == 2) {
      if (strcmp("ms", suffix) == 0) {
         ret = quantum_lookup_msec(value, quantum_ptr);
         if (ret == QUANTUM_FAILURE) {
            fatal("unsupported quantum; "
                  "see --quantum=list for supported values");
         } else {
            return QUANTUM_SUCCESS;
         }
      } else {
         fatal("invalid quantum specifier: %s", arg);
      }
   }
   return QUANTUM_FAILURE;
}

/*
 *  Dump the quantum map without much fanfare.
 */
static void print_quantum_map(void) {
   int i;
   int count = sizeof(quantum_map)/sizeof(quantum_entry_t);

   init_quantum_map();

   for (i=0; i < count; i++) {
      printf("%3d => %3dms\n", quantum_map[i].nice, quantum_map[i].msec);
   }
}


/***********************************************************************
 *                                                                     *
 *  Option processing routines                                         *
 *                                                                     *
 ***********************************************************************/


static void process_bias_option(const char* arg, cpuset_t* setp)
{
   int_list_t* node;

   cpuset_init(setp);

   if (strcmp(arg, "active") == 0) {
      init_valid_cpu_set();
      cpuset_copy(setp, valid_cpu_setp);
   } else if (strcmp(arg, "boot") == 0) {
      int ret;

      ret = mpadvise(MPA_CPU_BOOT, 0, 0, setp);
      if (ret == MPA_SUCCESS) {
         fatal("cannot determine boot cpu");
      }
   } else {
      node = parse_int_list(arg, "bias");
      while (node) {
         validate_cpu(node->data);
         cpuset_set_cpu(setp, node->data, 1);
         node = node->next;
      }
   }
}


static int process_policy_option(const char* arg)
{
   const char* str = arg;

   /* make SCHED_ optional */
   if (strncasecmp("SCHED_", str, 6) == 0) {
      str += 6;
   }

   if (strcasecmp(str, "OTHER") == 0) {
      return SCHED_OTHER;
   } else if (strcasecmp(str, "RR") == 0) {
      return SCHED_RR;
   } else if (strcasecmp(str, "FIFO") == 0) {
      return SCHED_FIFO;
   } else {
      fatal("unknown scheduling policy: %s", arg);
      return -1;
   }
}


static int process_priority_option(const char* arg)
{
   int pri;

   pri = strtol(arg, NULL, 10);
   if (pri < 0) {
      fatal("invalid negative priority");
   }
   return pri;
}


/*
 *  The --user option must accept uids and user names.
 */
static int_list_t* process_user_option(const char* arg)
{
   string_list_t* user_list;
   string_list_t* node;
   int_list_t* uid_list_ptr = NULL;
   uid_t uid;
   struct passwd* pwent;

   user_list = parse_string_list(arg);

   node = user_list;
   while (node) {
      if (is_only_digits(node->data)) {
         uid = strtoul(node->data, NULL, 10);
         pwent = getpwuid(uid);
         if (pwent == NULL) {
            fatal("unknown uid: %d", uid);
         }
         insert_int(&uid_list_ptr, pwent->pw_uid);
      } else {
         pwent = getpwnam(node->data);
         if (pwent == NULL) {
            fatal("unknown user: %s", node->data);
         }
         insert_int(&uid_list_ptr, pwent->pw_uid);
      }    
      node = node->next;
   }

   if (uid_list_ptr == NULL) {
      fatal("no matching users found");
   }

   free_string_list(&user_list);
   return uid_list_ptr;
}


/*
 *  On Linux we need to bound quantum to valid nice values.
 *  We also accept quantum arguments of the form "200ms".
 *  In addition, a quantum argument of "list" shows the mapping
 *  of nice values to quantum times.
 */
static int process_quantum_option(char* arg)
{
   int quantum_val;
   int ret;

   if (strcmp("list", arg) == 0) {
      print_quantum_map();
      exit(0);
   }

   ret = quantum_check_msec(arg, &quantum_val);
   if (ret == QUANTUM_FAILURE) {
      quantum_val = strtol(arg, NULL, 10);
   }

   if (quantum_val < -20 || quantum_val > 19) {
      fatal("quantum must be between -20 and 19 (inclusive)");
   }
   return quantum_val;
}


static int process_copies_option(const char* arg)
{
   int num_copies;

   num_copies = strtol(arg, NULL, 10);
   if (num_copies <= 0) {
      fatal("invalid copies count");
   }
   if (num_copies > MAX_COPIES) {
      fatal("maximum of %d copies allowed", MAX_COPIES);
   }
   return num_copies;
}


/*
 *  Called when processing a process specifier to ensure that only
 *  one process specifier is specified.
 */
static void check_for_duplicate_specifier(void)
{
   if (opt_pid || opt_tid || opt_user || opt_group || opt_name || opt_all) {
      fatal("only one process/thread specifier option may be specified");
   }
}


static void process_options(int argc, char* argv[])
{
   int c;
   int i;

   while (1) {
      i = 0;
      c = getopt_long(argc, argv, short_options, long_options, &i);

      if (c == -1) {
         break;
      }

      switch (c) {
         case OPT_UNKNOWN: {
            exit(1);
            break;
         }

         case OPT_ARG_MISSING: {
            exit(1);
            break;
         }

         case OPT_BIAS: {
            if (opt_bias) {
               fatal("only one bias option may be specified");
            }
            opt_bias = 1;
            bias_cpu_setp = cpuset_alloc();
            if (!bias_cpu_setp) {
               fatal("out of memory");
            }
            process_bias_option(optarg, bias_cpu_setp);
            break;
         }

         case OPT_POLICY: {
            if (opt_policy) {
               fatal("only one policy option may be specified");
            }
            opt_policy = 1;
            policy = process_policy_option(optarg);
            break;
         }

         case OPT_PRIORITY: {
            if (opt_priority) {
               fatal("only one priority option may be specified");
            }
            opt_priority = 1;
            priority = process_priority_option(optarg);
            break;
         }

         case OPT_ALL: {
            check_for_duplicate_specifier();
            opt_all = 1;
            break;
         }

         case OPT_PID: {
            check_for_duplicate_specifier();
            opt_pid = 1;
            pid_list = parse_int_list(optarg, "pid");
            break;
         }

         case OPT_TID: 
         case OPT_LWPID: {
            check_for_duplicate_specifier();
            opt_tid = 1;
            tid_list = parse_int_list(optarg, "tid");
            break;
         }

         case OPT_USER: {
            check_for_duplicate_specifier();
            opt_user = 1;
            uid_list = process_user_option(optarg);
            break;
         }

         case OPT_GROUP: {
            check_for_duplicate_specifier();
            opt_group = 1;
            pgrp_list = parse_int_list(optarg, "group");
            break;
         }

         case OPT_NAME: {
            check_for_duplicate_specifier();
            opt_name = 1;
            name_list = parse_string_list(optarg);
            break;
         }

         case OPT_QUANTUM: {
            if (opt_quantum) {
               fatal("only one quantum option may be specified");
            }
            opt_quantum = 1;
            quantum_arg = process_quantum_option(optarg);
            break;
         }

         case OPT_COPIES: {
            if (opt_copies) {
               fatal("only one copies option may be specified");
            }
            opt_copies = 1;
            copies = process_copies_option(optarg);
            break;
         }

         case OPT_FORK: {
            if (opt_fork) {
               fatal("only one fork option may be specified");
            }
            opt_fork = 1;
            break;
         }

         case OPT_INFO: {
            if (opt_info) {
               fatal("only one info option may be specified");
            }
            opt_info = 1;
            break;
         }

         case OPT_NEGATE: {
            if (opt_negate) {
               fatal("only one fork option may be specified");
            }
            opt_negate = 1;
            break;
         }

         case OPT_VERBOSE: {
            if (opt_verbose) {
               fatal("only one verbose option may be specified");
            }
            opt_verbose = 1;
            break;
         }

         case OPT_VERSION: {
            print_version();
            break;
         }

         case OPT_HELP: {
            print_help();
            break;
         }

         case OPT_CPUS: {
            init_valid_cpu_set();
            printf("%d\n", cpuset_count(valid_cpu_setp));
            exit(0);
            break;
         }

         case OPT_DEBUG: {
            opt_debug = 1;
            break;
         }

         default: {
            internal_error();
            break;
         }
      }
   }

   /*
    *  Error check the specified arguments and options.
    */

   if (optind < argc) {
      /*
       *  A command has been specified.
       */
      if (opt_pid || opt_tid || opt_group || opt_user || opt_name || opt_all) {
         fatal("specify existing processes/threads or a command, not both");
      }
      if (opt_info) {
         fatal("must specify info option with existing processes/threads");
      }
      command = optind;
   } else {
      /*
       *  No command has been specified.
       */
      if (!opt_bias && !opt_policy && !opt_priority && !opt_quantum) {
         /*
          *  No action was specified, so assume the user just wants info.
          */
         opt_info = 1;
      }
      if (!opt_pid && !opt_tid && !opt_group && !opt_user && !opt_name && !opt_all) {
         /*
          *  No process specifier options were specified, so assume that
          *  the user wants info about the run process itself.
          */
         opt_info = 1;
         insert_int(&tid_list, gettid());
      };
   }

   if (opt_negate) {
      if (opt_bias) {
         cpuset_negate(bias_cpu_setp);
      } else {
         fatal("no bias specified");
      }
   }

   if (opt_fork && !command) {
      fatal("fork option only valid with command");
   }

   if (opt_copies && !command) {
      fatal("copies option only valid with command");
   }

   if (opt_all && (opt_policy || opt_priority || opt_bias || opt_quantum)
               && !is_ccur_kernel()) {
      /*
       *  Attempting to set scheduler environment for all existing processes
       *  with a non-RedHawk kernel.  This is generally a very bad thing to do
       *  without special added kernel protections.
       */
      fatal("not allowed to set scheduler parameters for all processes");
   }
}


/***********************************************************************
 *                                                                     *
 *  Proc walk wrapper                                                  *
 *                                                                     *
 ***********************************************************************/

static pwid_t* local_pwid;

/*
 *  Create a new walk but re-use an existing one if possible.
 */
static pwid_t* local_proc_walk_init(void)
{
   if (local_pwid) {
      proc_walk_reset(local_pwid);
   } else {
      local_pwid = proc_walk_init();
      if (local_pwid == NULL) {
         fatal("unable to walk /proc");
      }
   }

   return local_pwid;
}


static void local_proc_walk_done(void)
{
   if (local_pwid) {
      proc_walk_done(local_pwid);
      local_pwid = NULL;
   }
}


/***********************************************************************
 *                                                                     *
 *  Dump pid environment                                               *
 *                                                                     *
 ***********************************************************************/


static cpuset_t* unshielded_cpu_setp;
static int show_actual_bias;


static void dump_info_init(void)
{
   FILE* shielded_file;
   int ret;
   cpuset_t* shielded_cpu_setp;

   init_valid_cpu_set();

   /*
    *  Check to see if shielding is present and enabled and if so
    *  produce another set of CPU bias statistics taking shielding
    *  into account.
    */
   shielded_cpu_setp = cpuset_alloc();
   shielded_file = fopen("/proc/shield/procs", "r");
   if (shielded_file == NULL) {
      cpuset_init(shielded_cpu_setp);
      debug("assuming system does not support shielding");
   } else {
      char buf[RUN_BUFSIZ];
      ret = fread(buf, sizeof(char), RUN_BUFSIZ, shielded_file);
      if (ret == 0 || ret >= RUN_BUFSIZ) {
         warn("error reading processor shielding info");
         cpuset_init(shielded_cpu_setp);
      } else {
         cpuset_set_string(shielded_cpu_setp, buf);
         show_actual_bias = 1;
         debug("assuming system supports shielding");
      }
      fclose(shielded_file);
   }

   unshielded_cpu_setp = cpuset_alloc();
   cpuset_copy(unshielded_cpu_setp, shielded_cpu_setp);
   cpuset_negate(unshielded_cpu_setp);
   cpuset_and_eq(unshielded_cpu_setp, valid_cpu_setp);
   cpuset_free(shielded_cpu_setp);
}


static void dump_info_done(void)
{
   cpuset_free(unshielded_cpu_setp);
}


static void dump_tid_info(pid_t tid)
{
   static int first = 1;
   int ret;
   cpuset_t* bias_cpu_set_ptr;
   char* name = NULL;
   char* bias_string = NULL;
   char* actual_bias_string = NULL;
   int policy_val;
   const char* policy_string = NULL;
   struct sched_param param;
   int nice_val;
   pid_t pid;

   ret = proc_stat(tid, PROC_STAT_CMD, (void*)&name);
   if (ret == PROC_STAT_FAILURE || name == NULL) {
      if (opt_tid || opt_pid) {
         warn("can't stat %d", tid);
      } else {
         debug("can't stat %d", tid);
      }
      goto out;
   }

   ret = proc_stat(tid, PROC_STAT_PID, (void*)&pid);
   if (ret == PROC_STAT_FAILURE || pid == 0) {
      if (opt_tid || opt_pid) {
         warn("can't stat %d", tid);
      } else {
         debug("can't stat %d", tid);
      }
      goto out;
   }

   bias_cpu_set_ptr = cpuset_alloc();
   ret = mpadvise(MPA_PRC_GETBIAS, MPA_TID, tid, bias_cpu_set_ptr);
   if (ret == MPA_FAILURE) {
      debug("failed to determine bias for %d", tid);
      if (errno == ENOSYS && first) {
         debug("is sched_getaffinity supported in kernel?");
      }
      /* no affinity info so assume full cpu set */
      cpuset_copy(bias_cpu_set_ptr, valid_cpu_setp);
   }
   cpuset_and_eq(bias_cpu_set_ptr, valid_cpu_setp);
   bias_string = cpuset_get_string(bias_cpu_set_ptr);
   if (show_actual_bias) {
      cpuset_and_eq(bias_cpu_set_ptr, unshielded_cpu_setp);
      /*
       *  The actual bias is the process bias with shielded cpus removed.
       *  If the process is biased to only shielded cpus, the actual bias
       *  is the same as the process bias.
       */
      if (cpuset_count(bias_cpu_set_ptr) != 0) {
         actual_bias_string = cpuset_get_string(bias_cpu_set_ptr);
      } else {
         actual_bias_string = strdup(bias_string);
         debug("process is exclusively biased\n");
      }
   }
   cpuset_free(bias_cpu_set_ptr);

   policy_val = sched_getscheduler(tid);
   if (policy_val == -1) {
      debug("can't determine scheduler policy for %d", tid);
      goto out;
   }
   if (policy_val == SCHED_OTHER) {
      policy_string = "other";
   } else if (policy_val == SCHED_FIFO) {
      policy_string = "fifo";
   } else if (policy_val == SCHED_RR) {
      policy_string = "rr";
   } else {
      fatal("invalid scheduler policy for %d", tid);
   }

   ret = sched_getparam(tid, &param);
   if (ret == -1) {
      debug("can't get scheduler parameters for %d", tid);
      goto out;
   }

   errno = 0;
   nice_val = getpriority(PRIO_PROCESS, tid);
   if (errno) {
      debug("unable to determine priority for id %d", tid);
      goto out;
   }

   if (first) {
      first = 0;
      if (show_actual_bias) {
         printf("Pid    Tid    Bias   Actual   Policy   Pri   Nice   Name\n");
      } else {
         printf("Pid    Tid    Bias   Policy   Pri   Nice   Name\n");
      }
   }

   printf("%-6d ", pid);
   printf("%-6d ", tid);
   printf("0x%-4s ", bias_string);
   if (show_actual_bias) {
      printf("0x%-4s   ", actual_bias_string);
   }
   printf("%-5s    ", policy_string);
   printf("%-4d  ", param.sched_priority);
   printf("%-4d   ", nice_val);
   printf("%s\n", name);

out:
   if (name) {
      free(name);
   }
   if (bias_string) {
      free(bias_string);
   }
   if (show_actual_bias && actual_bias_string) {
      free(actual_bias_string);
   }
}


/*
 *  Reverse the list (necessary because lists only prepend).
 */
static void dump_info_list(int_list_t* node)
{
    if (node) {
       dump_info_list(node->next);
       dump_tid_info(node->data);
    }
}

/***********************************************************************
 *                                                                     *
 *  The main routine                                                   *
 *                                                                     *
 ***********************************************************************/


int main(int argc, char* argv[])
{
   int ret;

   set_prog_name(argv[0]);
   process_options(argc, argv);

   if (command) {
      if (opt_fork || copies > 1) {
         pid_t pid = -1;
         int copy;

         for (copy = 0; copy < copies; copy++) {
            pid = fork();
            if (pid == -1) {
               fatal("unable to fork");
            }
            if (pid == 0) {
               if (opt_verbose) {
                  info("pid %d created", getpid());
               }
               break;
            }
         }
         if (pid) {
            if (opt_fork) {
               /*
                *  This is the end of the road for the parent.
                */
               if (opt_verbose) {
                  info("exiting");
               }
               return 0;
            } else {
               /*
                *  Wait around for all the children to exit.
                *  We only get here if multiple copies requested.
                */
               while (1) {
                  pid_t child;

                  child = wait(NULL);
                  if (child == -1) {
                     if (errno == ECHILD) {
                        if (opt_verbose) {
                           info("exiting");
                        }
                        return 0;
                     } else {
                        internal_error();
                     }
                  } else if (child) {
                     if (opt_verbose) {
                        info("pid %d exited", child);
                     }
                  }
               }
            }
         }
      }
      /*
       *  To simplify the code, pretend that the user specified
       *  a --tid option with the current (or child) tid.
       */
      insert_int(&tid_list, gettid());
   }

   /*
    *  Note that if a command was forked, then the child is
    *  executing the following code.
    */

   if (pid_list) {
      /*
       *  To simplify the code, turn the list of process IDs
       *  into a --tid option with all matching tids.
       */
      int_list_t* node;
      twid_t* twid;
      pid_t tid;

      for (node = pid_list; node != NULL; node = node->next) {
         twid = thread_walk_init(node->data);
         if (twid == NULL) {
            debug("can't find pid %d", node->data);
            continue;
         }
         while ((tid = thread_walk_get_next_tid(twid))) {
            insert_int(&tid_list, tid);
         }
         thread_walk_done(twid);
      }

      if (tid_list == NULL) {
         fatal("no matching processes/threads found");
      }
   }

   if (name_list) {
      /*
       *  To simplify the code, turn the list of process names
       *  into a --tid option with all matching tids.
       */
      string_list_t* node;
      pid_t tid;
      int code;
      char* name;
      pwid_t* pwid;

      pwid = local_proc_walk_init();
      while ((tid = proc_walk_get_next_tid(pwid))) {
         name = NULL;
         code = proc_stat(tid, PROC_STAT_CMD, (void*)&name);
         if (code == PROC_STAT_FAILURE || name == NULL) {
            debug("can't stat tid %d", tid);
            continue;
         }
         node = name_list;
         while (node) {
            if (strcmp(name, node->data) == 0) {
               insert_int(&tid_list, tid);
               break;
            }
            node = node->next;
         }
         free(name);
      }

      if (tid_list == NULL) {
         fatal("no matching processes/threads found");
      }
   }

   if (opt_all) {
      /*
       *  To simplify the code, create a tid_list with all existing tids.
       */
      pid_t tid;
      pwid_t* pwid;

      pwid = local_proc_walk_init();
      while ((tid = proc_walk_get_next_tid(pwid))) {
         insert_int(&tid_list, tid);
      }

      if (tid_list == NULL) {
         debug("no existing processes/threads found!");
         internal_error();
      }
   }

   /*
    *  Process bias adjustements using mpadvise(3).
    */
   if (opt_bias) {
      int_list_t* node = NULL;
      int type = 0;
      const char* name = NULL;

      if (tid_list) {
         node = tid_list;
         type = MPA_TID;
         name = "thread";
      } else if (pgrp_list) {
         node = pgrp_list;
         type = MPA_PGID;
         name = "process group";
      } else if (uid_list) {
         node = uid_list;
         type = MPA_UID;
         name = "user id";
      } else {
         internal_error();
      }

      while (node) {
         if (opt_verbose) {
            info("setting bias for %s %d", name, node->data);
         }
         ret = mpadvise(MPA_PRC_SETBIAS, type, node->data, bias_cpu_setp);
         if (ret == MPA_FAILURE) {
            switch (errno) {
               case ESRCH: {
                  warn("%s %d not found", name, node->data);
                  break;
               }
               case EACCES: {
                  warn("operation not permitted for %s %d", name, node->data);
                  break;
               }
               default: {
                  warn("failed to set bias for %s %d", name, node->data);
                  break;
               }
            }
         }
         node = node->next;
      }
   }

   /*
    *  Process quantum adjustments using setpriority(2).
    */
   if (opt_quantum) {
      int_list_t* node = NULL;
      int type = 0;
      const char* name = NULL;

      if (tid_list) {
         node = tid_list;
         type = PRIO_PROCESS;
         name = "thread";
      } else if (pgrp_list) {
         node = pgrp_list;
         type = PRIO_PGRP;
         name = "process group";
      } else if (uid_list) {
         node = uid_list;
         type = PRIO_USER;
         name = "user id";
      } else {
         internal_error();
      }

      while (node) {
         if (opt_verbose) {
            info("setting quantum for %s %d", name, node->data);
         }
         errno = 0;
         setpriority(type, node->data, quantum_arg);
         switch (errno) {
            case 0: {
               break;
            }
            case ESRCH: {
               warn("%s %d not found", name, node->data);
               break;
            }
            case EINVAL: {
               warn("invalid quantum specified for %s %d", name, node->data);
               break;
            }
            case EPERM: case EACCES: {
               warn("operation not permitted for %s %d", name, node->data);
               break;
            }
            default: {
               warn("failed to set quantum for %s %d", name, node->data);
               break;
            }
         }
         node = node->next;
      }
   }

   if (opt_group && (opt_policy || opt_priority || opt_info)) {
      /*
       *  Turn the list of process groups into a --tid option
       *  with all matching tids.  This is needed because
       *  sched_setparam(2) only operates on tids.
       */
      int_list_t* group = pgrp_list;

      while (group) {
         pid_t tid;
         pid_t pgrp;
         pwid_t* pwid;

         pwid = local_proc_walk_init();
         while ((tid = proc_walk_get_next_tid(pwid))) {
            pgrp = -1;
            ret = proc_stat(tid, PROC_STAT_PGRP, &pgrp);
            if (ret == PROC_STAT_FAILURE || pgrp == -1) {
               debug("can't stat tid %d", tid);
               continue;
            }
            if (pgrp == group->data) {
               insert_int(&tid_list, tid);
            }
         }
         group = group->next;
      }

      if (tid_list == NULL) {
         fatal("no matching processes/threads found");
      }
   }

   if (opt_user && (opt_policy || opt_priority || opt_info)) {
      /*
       *  Turn the list of uids into a --tid option
       *  with all matching tids.  This is needed because
       *  sched_setparam(2) only operates on tids.
       */
      int_list_t* user = uid_list;

      while (user) {
         pid_t tid;
         uid_t uid;
         pwid_t* pwid;

         pwid = local_proc_walk_init();
         while ((tid = proc_walk_get_next_tid(pwid))) {
            uid = -1;
            ret = proc_stat(tid, PROC_STAT_UID, &uid);
            if (ret == PROC_STAT_FAILURE || (int)uid == -1) {
               debug("unable to open /proc/%d/stat", tid);
               continue;
            }
            if ((int)uid == user->data) {
               insert_int(&tid_list, tid);
            }
         }
         user = user->next;
      }

      if (tid_list == NULL) {
         fatal("no matching processes/threads found");
      }
   }

   local_proc_walk_done();

   /*
    *  We have three cases here:
    *
    *  opt_policy && opt_priority
    *     Call sched_setscheduler to set policy and priority.
    *
    *  opt_policy && !opt_priority
    *     Default to priority 0.
    *     Call sched_setscheduler to set policy and priority.
    *
    *  !opt_policy && opt_priority
    *     Call sched_setparam to set priority.
    */

   if (opt_policy) {
      int_list_t* node = tid_list;
      struct sched_param param;
   
      if (!opt_priority) {
         priority = 0;
      }

      param.sched_priority = priority;
      while (node) {
         if (opt_verbose) {
            info("setting policy/priority for process %d", node->data);
         }
         ret = sched_setscheduler(node->data, policy, &param);
         if (ret == -1) {
            switch (errno) {
               case ESRCH: {
                  warn("process not found: %d", node->data);
                  break;
               }
               case EPERM: {
                  warn("operation not permitted for process %d", node->data);
                  break;
               }
               case EINVAL: {
                  if (priority == 0 && policy == SCHED_FIFO) {
                     warn("priority 0 disallowed for policy SCHED_FIFO");
                  } else if (priority > 99 && policy == SCHED_FIFO) {
                     warn("priority > 99 disallowed for policy SCHED_FIFO");
                  } else if (priority == 0 && policy == SCHED_RR) {
                     warn("priority 0 disallowed for policy SCHED_RR");
                  } else if (priority > 99 && policy == SCHED_RR) {
                     warn("priority > 99 disallowed for policy SCHED_RR");
                  } else if (priority != 0 && policy == SCHED_OTHER) {
                     warn("non-zero priority disallowed for policy SCHED_OTHER");
                  } else {
                     warn("invalid policy/priority specified for process %d", node->data);
                  }
                  break;
               }
               default: {
                  debug("unexpected error from sched_setscheduler");
                  perror("run");
                  warn("failed to set policy/priority for process %d", node->data);
                  break;
               }
            }
         }
         node = node->next;
      }
   } else if (opt_priority) {
      int_list_t* node = tid_list;
      struct sched_param param;

      param.sched_priority = priority;
      while (node) {
         if (opt_verbose) {
            info("setting priority for process %d", node->data);
         }
         ret = sched_setparam(node->data, &param);
         if (ret == -1) {
            switch (errno) {
               case ESRCH: {
                  warn("process not found: %d", node->data);
                  break;
               }
               case EPERM: {
                  warn("operation not permitted for process %d", node->data);
                  break;
               }
               case EINVAL: {
                  int pol = sched_getscheduler(node->data);  
                  if (priority == 0 && pol == SCHED_FIFO) {
                     warn("priority 0 disallowed for policy SCHED_FIFO");
                  } else if (priority > 99 && pol == SCHED_FIFO) {
                     warn("priority > 99 disallowed for policy SCHED_FIFO");
                  } else if (priority == 0 && pol == SCHED_RR) {
                     warn("priority 0 disallowed for policy SCHED_RR");
                  } else if (priority > 99 && pol == SCHED_RR) {
                     warn("priority > 99 disallowed for policy SCHED_RR");
                  } else if (priority != 0 && pol == SCHED_OTHER) {
                     warn("non-zero priority disallowed for policy SCHED_OTHER");
                  } else {
                     warn("invalid policy/priority specified for process %d", node->data);
                  }

                  warn("invalid priority specified for process %d", node->data);
                  break;
               }
               default: {
                  debug("unexpected error from sched_setparam");
                  perror("run");
                  warn("failed to set priority for process %d", node->data);
                  break;
               }
            }
         }
         node = node->next;
      }
   }

   if (opt_info) {
      dump_info_init();
      dump_info_list(tid_list);
      dump_info_done();
      exit(0);
   }

   if (tid_list) free_int_list(&tid_list);
   if (pid_list) free_int_list(&pid_list);
   if (uid_list) free_int_list(&uid_list);
   if (pgrp_list) free_int_list(&pgrp_list);
   if (name_list) free_string_list(&name_list);

   if (command) {
      execvp(argv[command], &argv[command]);
      /* child should not return */
      fatal("unable to exec");
   }

   debug("exiting with error count: %d", error_count);

   return (error_count != 0);
}
